import argparse
import sys


def cli():
    parser = argparse.ArgumentParser(description="Swarms CLI")
    parser.add_argument(
        "file_name", help="Python file containing Swarms code to run"
    )
    # Help message for the -h flag is automatically generated by argparse
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s 0.1.0"
    )

    # Check deployments for a given model
    parser.add_argument(
        "-c", "--check", help="Check deployments for a given agent"
    )

    # Generate an API key for a given agent
    parser.add_argument(
        "-g",
        "--generate",
        help="Generate an API key for a given agent",
    )

    # Signin to swarms with a given API key
    parser.add_argument(
        "-s", "--signin", help="Signin to swarms with a given API key"
    )

    # Signout of swarms
    parser.add_argument("-o", "--signout", help="Signout of swarms")

    # List all agents
    parser.add_argument("-l", "--list", help="List all agents")

    # List all deployments
    parser.add_argument(
        "-d", "--deployments", help="List all deployments"
    )

    # Pricing information
    parser.add_argument("-p", "--pricing", help="Pricing information")

    # Run a deployment
    parser.add_argument("-r", "--run", help="Run a deployment")

    # Stop a deployment
    parser.add_argument("-t", "--stop", help="Stop a deployment")

    # Delete a deployment
    parser.add_argument("-x", "--delete", help="Delete a deployment")

    # Get a deployment
    parser.add_argument("-e", "--get", help="Get a deployment")

    # Get a deployment's logs
    parser.add_argument(
        "-z", "--logs", help="Get a deployment's logs"
    )

    # Parse the arguments
    args = parser.parse_args()

    # Execute the specified file
    try:
        with open(args.file_name, "r") as file:
            exec(file.read(), globals())
    except FileNotFoundError:
        print(f"Error: File '{args.file_name}' not found.")
        sys.exit(1)
    except Exception as e:
        print(f"Error executing file '{args.file_name}': {e}")
        sys.exit(1)



def presidential_debate(character_names, topic):
    game_description = f"""Here is the topic for the presidential debate: {topic}.
    The presidential candidates are: {', '.join(character_names)}."""

    return game_description


def character(character_name, topic, word_limit):
    prompt = f"""
    You will speak in the style of {character_name}, and exaggerate their personality.
    You will come up with creative ideas related to {topic}.
    Do not say the same things over and over again.
    Speak in the first person from the perspective of {character_name}
    For describing your own body movements, wrap your description in '*'.
    Do not change roles!
    Do not speak from the perspective of anyone else.
    Speak only from the perspective of {character_name}.
    Stop speaking the moment you finish speaking from your perspective.
    Never forget to keep your response to {word_limit} words!
    Do not add anything else.
    """
    return prompt


def debate_monitor(game_description, word_limit, character_names):
    prompt = f"""

    {game_description}
    You are the debate moderator.
    Please make the debate topic more specific.
    Frame the debate topic as a problem to be solved.
    Be creative and imaginative.
    Please reply with the specified topic in {word_limit} words or less.
    Speak directly to the presidential candidates: {*character_names,}.
    Do not add anything else.
    """

    return prompt


def generate_character_header(
    game_description, topic, character_name, character_description
):
    pass



from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline


class MultiModalLlava:
    """
    LLava Model

    Args:
        model_name_or_path: The model name or path to the model
        revision: The revision of the model to use
        device: The device to run the model on
        max_new_tokens: The maximum number of tokens to generate
        do_sample: Whether or not to use sampling
        temperature: The temperature of the sampling
        top_p: The top p value for sampling
        top_k: The top k value for sampling
        repetition_penalty: The repetition penalty for sampling
        device_map: The device map to use

    Methods:
        __call__: Call the model
        chat: Interactive chat in terminal

    Example:
        >>> from swarms.models.llava import LlavaModel
        >>> model = LlavaModel(device="cpu")
        >>> model("Hello, I am a robot.")
    """

    def __init__(
        self,
        model_name_or_path="TheBloke/llava-v1.5-13B-GPTQ",
        revision="main",
        device="cuda",
        max_new_tokens=512,
        do_sample=True,
        temperature=0.7,
        top_p=0.95,
        top_k=40,
        repetition_penalty=1.1,
        device_map: str = "auto",
    ):
        self.device = device
        self.model = AutoModelForCausalLM.from_pretrained(
            model_name_or_path,
            device_map=device_map,
            trust_remote_code=False,
            revision=revision,
        ).to(self.device)

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, use_fast=True
        )
        self.pipe = pipeline(
            "text-generation",
            model=self.model,
            tokenizer=self.tokenizer,
            max_new_tokens=max_new_tokens,
            do_sample=do_sample,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            repetition_penalty=repetition_penalty,
            device=0 if self.device == "cuda" else -1,
        )

    def __call__(self, prompt):
        """Call the model"""
        return self.pipe(prompt)[0]["generated_text"]

    def chat(self):
        """Interactive chat in terminal"""
        print(
            "Starting chat with LlavaModel. Type 'exit' to end the"
            " session."
        )
        while True:
            user_input = input("You: ")
            if user_input.lower() == "exit":
                break
            response = self(user_input)
            print(f"Model: {response}")



from swarms.structs.agent import Agent
from swarms.structs.autoscaler import AutoScaler
from swarms.structs.base import BaseStructure
from swarms.structs.base_swarm import AbstractSwarm
from swarms.structs.base_workflow import BaseWorkflow
from swarms.structs.concurrent_workflow import ConcurrentWorkflow
from swarms.structs.conversation import Conversation
from swarms.structs.groupchat import GroupChat, GroupChatManager
from swarms.structs.model_parallizer import ModelParallelizer
from swarms.structs.multi_agent_collab import MultiAgentCollaboration
from swarms.structs.nonlinear_workflow import NonlinearWorkflow
from swarms.structs.recursive_workflow import RecursiveWorkflow
from swarms.structs.schemas import (
    Artifact,
    ArtifactUpload,
    StepInput,
    TaskInput,
)
from swarms.structs.sequential_workflow import SequentialWorkflow
from swarms.structs.swarm_net import SwarmNetwork
from swarms.structs.utils import (
    distribute_tasks,
    extract_key_from_json,
    extract_tokens_from_text,
    find_agent_by_id,
    find_token_in_text,
    parse_tasks,
    detect_markdown,
)
from swarms.structs.task import Task
from swarms.structs.block_wrapper import block

__all__ = [
    "Agent",
    "SequentialWorkflow",
    "AutoScaler",
    "Conversation",
    "TaskInput",
    "Artifact",
    "ArtifactUpload",
    "StepInput",
    "SwarmNetwork",
    "ModelParallelizer",
    "MultiAgentCollaboration",
    "AbstractSwarm",
    "GroupChat",
    "GroupChatManager",
    "parse_tasks",
    "find_agent_by_id",
    "distribute_tasks",
    "find_token_in_text",
    "extract_key_from_json",
    "extract_tokens_from_text",
    "ConcurrentWorkflow",
    "RecursiveWorkflow",
    "NonlinearWorkflow",
    "BaseWorkflow",
    "BaseStructure",
    "detect_markdown",
    "Task",
    "block",
]



from swarms.memory.base_vectordb import VectorDatabase
from swarms.memory.short_term_memory import ShortTermMemory
from swarms.memory.sqlite import SQLiteDB
from swarms.memory.weaviate_db import WeaviateDB

__all__ = [
    "VectorDatabase",
    "ShortTermMemory",
    "SQLiteDB",
    "WeaviateDB",
]



import platform
import subprocess

import pkg_resources
import psutil
import toml


def get_python_version():
    return platform.python_version()


def get_pip_version():
    try:
        pip_version = (
            subprocess.check_output(["pip", "--version"])
            .decode()
            .split()[1]
        )
    except Exception as e:
        pip_version = str(e)
    return pip_version


def get_oi_version():
    try:
        oi_version_cmd = (
            subprocess.check_output(["interpreter", "--version"])
            .decode()
            .split()[1]
        )
    except Exception as e:
        oi_version_cmd = str(e)
    oi_version_pkg = pkg_resources.get_distribution(
        "open-interpreter"
    ).version
    oi_version = oi_version_cmd, oi_version_pkg
    return oi_version


def get_os_version():
    return platform.platform()


def get_cpu_info():
    return platform.processor()


def get_ram_info():
    vm = psutil.virtual_memory()
    used_ram_gb = vm.used / (1024**3)
    free_ram_gb = vm.free / (1024**3)
    total_ram_gb = vm.total / (1024**3)
    return (
        f"{total_ram_gb:.2f} GB, used: {used_ram_gb:.2f}, free:"
        f" {free_ram_gb:.2f}"
    )


def get_package_mismatches(file_path="pyproject.toml"):
    with open(file_path, "r") as file:
        pyproject = toml.load(file)
    dependencies = pyproject["tool"]["poetry"]["dependencies"]
    dev_dependencies = pyproject["tool"]["poetry"]["group"]["dev"][
        "dependencies"
    ]
    dependencies.update(dev_dependencies)

    installed_packages = {
        pkg.key: pkg.version for pkg in pkg_resources.working_set
    }

    mismatches = []
    for package, version_info in dependencies.items():
        if isinstance(version_info, dict):
            version_info = version_info["version"]
        installed_version = installed_packages.get(package)
        if installed_version and version_info.startswith("^"):
            expected_version = version_info[1:]
            if not installed_version.startswith(expected_version):
                mismatches.append(
                    f"\t  {package}: Mismatch,"
                    f" pyproject.toml={expected_version},"
                    f" pip={installed_version}"
                )
        else:
            mismatches.append(f"\t  {package}: Not found in pip list")

    return "\n" + "\n".join(mismatches)


def interpreter_info(interpreter):
    try:
        if interpreter.offline and interpreter.llm.api_base:
            try:
                curl = subprocess.check_output(
                    f"curl {interpreter.llm.api_base}"
                )
            except Exception as e:
                curl = str(e)
        else:
            curl = "Not local"

        messages_to_display = []
        for message in interpreter.messages:
            message = message.copy()
            try:
                if len(message["content"]) > 600:
                    message["content"] = (
                        message["content"][:300]
                        + "..."
                        + message["content"][-300:]
                    )
            except Exception as e:
                print(str(e), "for message:", message)
            messages_to_display.append(message)

        return f"""

        # Interpreter Info
        
        Vision: {interpreter.llm.supports_vision}
        Model: {interpreter.llm.model}
        Function calling: {interpreter.llm.supports_functions}
        Context window: {interpreter.llm.context_window}
        Max tokens: {interpreter.llm.max_tokens}

        Auto run: {interpreter.auto_run}
        API base: {interpreter.llm.api_base}
        Offline: {interpreter.offline}

        Curl output: {curl}

        # Messages

        System Message: {interpreter.system_message}

        """ + "\n\n".join([str(m) for m in messages_to_display])
    except:
        return "Error, couldn't get interpreter info"


def system_info(interpreter):
    oi_version = get_oi_version()
    print(f"""
        Python Version: {get_python_version()}
        Pip Version: {get_pip_version()}
        Open-interpreter Version: cmd:{oi_version[0]}, pkg: {oi_version[1]}
        OS Version and Architecture: {get_os_version()}
        CPU Info: {get_cpu_info()}
        RAM Info: {get_ram_info()}
        {interpreter_info(interpreter)}
    """)

    # Removed the following, as it causes `FileNotFoundError: [Errno 2] No such file or directory: 'pyproject.toml'`` on prod
    # (i think it works on dev, but on prod the pyproject.toml will not be in the cwd. might not be accessible at all)
    # Package Version Mismatches:
    # {get_package_mismatches()}



import logging

logger = logging.getLogger()
formatter = logging.Formatter("%(message)s")

ch = logging.StreamHandler()

ch.setFormatter(formatter)

logger.addHandler(ch)

logger.setLevel(logging.DEBUG)



from swarms.structs.conversation import Conversation
from swarms.models.base_llm import AbstractLLM


# Run the language model in a loop for n iterations
def SimpleAgent(
    llm: AbstractLLM = None, iters: int = 10, *args, **kwargs
):
    """Simple agent conversation

    Args:
        llm (_type_): _description_
        iters (int, optional): _description_. Defaults to 10.

    Example:
        >>> from swarms.models import GPT2LM
        >>> from swarms.agents import SimpleAgent
        >>> llm = GPT2LM()
        >>> SimpleAgent(llm, iters=10)
    """
    try:
        conv = Conversation(*args, **kwargs)
        for i in range(iters):
            user_input = input("User: ")
            conv.add("user", user_input)
            if user_input.lower() == "quit":
                break
            task = (
                conv.return_history_as_string()
            )  # Get the conversation history
            out = llm(task)
            conv.add("assistant", out)
            print(
                f"Assistant: {out}",
            )
            conv.display_conversation()
            conv.export_conversation("conversation.txt")

    except Exception as error:
        print(f"[ERROR][SimpleAgentConversation] {error}")
        raise error

    except KeyboardInterrupt:
        print("[INFO][SimpleAgentConversation] Keyboard interrupt")
        conv.export_conversation("conversation.txt")
        raise KeyboardInterrupt



# LLMs
from swarms.models.base_llm import AbstractLLM  # noqa: E402
from swarms.models.anthropic import Anthropic  # noqa: E402
from swarms.models.petals import Petals  # noqa: E402
from swarms.models.mistral import Mistral  # noqa: E402
from swarms.models.openai_models import (
    OpenAI,
    AzureOpenAI,
    OpenAIChat,
)  # noqa: E402

# from swarms.models.vllm import vLLM  # noqa: E402
from swarms.models.zephyr import Zephyr  # noqa: E402
from swarms.models.biogpt import BioGPT  # noqa: E402
from swarms.models.huggingface import HuggingfaceLLM  # noqa: E402
from swarms.models.wizard_storytelling import (
    WizardLLMStoryTeller,
)  # noqa: E402
from swarms.models.mpt import MPT7B  # noqa: E402
from swarms.models.mixtral import Mixtral  # noqa: E402

################# MultiModal Models
from swarms.models.base_multimodal_model import (
    BaseMultiModalModel,
)  # noqa: E402
from swarms.models.idefics import Idefics  # noqa: E402
from swarms.models.vilt import Vilt  # noqa: E402
from swarms.models.nougat import Nougat  # noqa: E402
from swarms.models.layoutlm_document_qa import (
    LayoutLMDocumentQA,
)  # noqa: E402
from swarms.models.gpt4_vision_api import GPT4VisionAPI  # noqa: E402
from swarms.models.openai_tts import OpenAITTS  # noqa: E402
from swarms.models.gemini import Gemini  # noqa: E402
from swarms.models.gigabind import Gigabind  # noqa: E402
from swarms.models.zeroscope import ZeroscopeTTV  # noqa: E402

# from swarms.models.gpt4v import GPT4Vision
# from swarms.models.dalle3 import Dalle3
# from swarms.models.distilled_whisperx import DistilWhisperModel # noqa: E402
# from swarms.models.whisperx_model import WhisperX  # noqa: E402
# from swarms.models.kosmos_two import Kosmos  # noqa: E402
# from swarms.models.cog_agent import CogAgent  # noqa: E402

from swarms.models.types import (
    TextModality,
    ImageModality,
    AudioModality,
    VideoModality,
    MultimodalData,
)  # noqa: E402

__all__ = [
    "AbstractLLM",
    "Anthropic",
    "Petals",
    "Mistral",
    "OpenAI",
    "AzureOpenAI",
    "OpenAIChat",
    "Zephyr",
    "BaseMultiModalModel",
    "Idefics",
    # "Kosmos",
    "Vilt",
    "Nougat",
    "LayoutLMDocumentQA",
    "BioGPT",
    "HuggingfaceLLM",
    "MPT7B",
    "WizardLLMStoryTeller",
    # "GPT4Vision",
    # "Dalle3",
    # "DistilWhisperModel",
    "GPT4VisionAPI",
    # "vLLM",
    "OpenAITTS",
    "Gemini",
    "Gigabind",
    "Mixtral",
    "ZeroscopeTTV",
    "TextModality",
    "ImageModality",
    "AudioModality",
    "VideoModality",
    "MultimodalData",
    # "CogAgent"
]



import tempfile
from enum import Enum
from typing import Any, Dict, Union

from langchain.utils import get_from_dict_or_env
from pydantic import root_validator

from swarms.tools.tool import BaseTool


def _import_elevenlabs() -> Any:
    try:
        import elevenlabs
    except ImportError as e:
        raise ImportError(
            "Cannot import elevenlabs, please install `pip install"
            " elevenlabs`."
        ) from e
    return elevenlabs


class ElevenLabsModel(str, Enum):
    """Models available for Eleven Labs Text2Speech."""

    MULTI_LINGUAL = "eleven_multilingual_v1"
    MONO_LINGUAL = "eleven_monolingual_v1"


class ElevenLabsText2SpeechTool(BaseTool):
    """Tool that queries the Eleven Labs Text2Speech API.

    In order to set this up, follow instructions at:
    https://docs.elevenlabs.io/welcome/introduction

    Attributes:
        model (ElevenLabsModel): The model to use for text to speech.
            Defaults to ElevenLabsModel.MULTI_LINGUAL.
        name (str): The name of the tool. Defaults to "eleven_labs_text2speech".
        description (str): The description of the tool.
            Defaults to "A wrapper around Eleven Labs Text2Speech. Useful for when you need to convert text to speech. It supports multiple languages, including English, German, Polish, Spanish, Italian, French, Portuguese, and Hindi."


    Usage:
    >>> from swarms.models import ElevenLabsText2SpeechTool
    >>> stt = ElevenLabsText2SpeechTool()
    >>> speech_file = stt.run("Hello world!")
    >>> stt.play(speech_file)
    >>> stt.stream_speech("Hello world!")

    """

    model: Union[ElevenLabsModel, str] = ElevenLabsModel.MULTI_LINGUAL

    name: str = "eleven_labs_text2speech"
    description: str = (
        "A wrapper around Eleven Labs Text2Speech. Useful for when"
        " you need to convert text to speech. It supports multiple"
        " languages, including English, German, Polish, Spanish,"
        " Italian, French, Portuguese, and Hindi. "
    )

    @root_validator(pre=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key exists in environment."""
        _ = get_from_dict_or_env(
            values, "eleven_api_key", "ELEVEN_API_KEY"
        )

        return values

    def _run(
        self,
        task: str,
    ) -> str:
        """Use the tool."""
        elevenlabs = _import_elevenlabs()
        try:
            speech = elevenlabs.generate(text=task, model=self.model)
            with tempfile.NamedTemporaryFile(
                mode="bx", suffix=".wav", delete=False
            ) as f:
                f.write(speech)
            return f.name
        except Exception as e:
            raise RuntimeError(
                f"Error while running ElevenLabsText2SpeechTool: {e}"
            )

    def play(self, speech_file: str) -> None:
        """Play the text as speech."""
        elevenlabs = _import_elevenlabs()
        with open(speech_file, mode="rb") as f:
            speech = f.read()

        elevenlabs.play(speech)

    def stream_speech(self, query: str) -> None:
        """Stream the text as speech as it is generated.
        Play the text in your speakers."""
        elevenlabs = _import_elevenlabs()
        speech_stream = elevenlabs.generate(
            text=query, model=self.model, stream=True
        )
        elevenlabs.stream(speech_stream)

    def save(self, speech_file: str, path: str) -> None:
        """Save the speech file to a path."""
        raise NotImplementedError(
            "Saving not implemented for this tool."
        )

    def __str__(self):
        return "ElevenLabsText2SpeechTool"



from dataclasses import dataclass


@dataclass
class BlockDevice:
    device: str
    cluster: str
    description: str

    def __init__(self, device: str, cluster: str, description: str):
        self.device = device
        self.cluster = cluster
        self.description = description

    def __str__(self):
        return (
            f"BlockDevice(device={self.device},"
            f" cluster={self.cluster},"
            f" description={self.description})"
        )



from concurrent import futures
from concurrent.futures import Future
from typing import TypeVar, Dict

T = TypeVar("T")


def execute_futures_dict(
    fs_dict: Dict[str, Future[T]]
) -> Dict[str, T]:
    """Execute a dictionary of futures and return the results.

    Args:
        fs_dict (dict[str, futures.Future[T]]): _description_

    Returns:
        dict[str, T]: _description_

    Example:
    >>> import concurrent.futures
    >>> import time
    >>> import random
    >>> import swarms.utils.futures
    >>> def f(x):
    ...     time.sleep(random.random())
    ...     return x
    >>> with concurrent.futures.ThreadPoolExecutor() as executor:
    ...     fs_dict = {
    ...         str(i): executor.submit(f, i)
    ...         for i in range(10)
    ...     }
    ...     print(swarms.utils.futures.execute_futures_dict(fs_dict))
    {'0': 0, '1': 1, '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7, '8': 8, '9': 9}

    """
    futures.wait(
        fs_dict.values(),
        timeout=None,
        return_when=futures.ALL_COMPLETED,
    )

    return {key: future.result() for key, future in fs_dict.items()}



import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Callable, List

from tabulate import tabulate
from termcolor import colored

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ModelParallelizer:
    """
    ModelParallelizer
    -----

    Architecture:
    How it works:
    1. ModelParallelizer receives a task from the user.
    2. ModelParallelizer distributes the task to all LLMs.
    3. ModelParallelizer collects the responses from all LLMs.
    4. ModelParallelizer prints the responses from all LLMs.

    Parameters:
    llms: list of LLMs

    Methods:
    run(task): distribute task to all LLMs and collect responses
    print_responses(task): print responses from all LLMs

    Usage:
    parallelizer = ModelParallelizer(llms)
    parallelizer.run(task)
    parallelizer.print_responses(task)


    """

    def __init__(
        self,
        llms: List[Callable] = None,
        load_balancing: bool = False,
        retry_attempts: int = 3,
        iters: int = None,
        *args,
        **kwargs,
    ):
        self.llms = llms
        self.load_balancing = load_balancing
        self.retry_attempts = retry_attempts
        self.iters = iters
        self.last_responses = None
        self.task_history = []

    def run(self, task: str):
        """Run the task string"""
        try:
            for i in range(self.iters):
                with ThreadPoolExecutor() as executor:
                    responses = executor.map(
                        lambda llm: llm(task), self.llms
                    )
                return list(responses)
        except Exception as error:
            print(
                f"[ERROR][ModelParallelizer] [ROOT CAUSE] [{error}]"
            )

    def print_responses(self, task):
        """Prints the responses in a tabular format"""
        responses = self.run_all(task)
        table = []
        for i, response in enumerate(responses):
            table.append([f"LLM {i+1}", response])
        print(
            colored(
                tabulate(
                    table,
                    headers=["LLM", "Response"],
                    tablefmt="pretty",
                ),
                "cyan",
            )
        )

    def run_all(self, task):
        """Run the task on all LLMs"""
        responses = []
        for llm in self.llms:
            responses.append(llm(task))
        return responses

    def print_arun_all(self, task):
        """Prints the responses in a tabular format"""
        responses = self.arun_all(task)
        table = []
        for i, response in enumerate(responses):
            table.append([f"LLM {i+1}", response])
        print(
            colored(
                tabulate(
                    table,
                    headers=["LLM", "Response"],
                    tablefmt="pretty",
                ),
                "cyan",
            )
        )

    # New Features
    def save_responses_to_file(self, filename):
        """Save responses to file"""
        with open(filename, "w") as file:
            table = [
                [f"LLM {i+1}", response]
                for i, response in enumerate(self.last_responses)
            ]
            file.write(tabulate(table, headers=["LLM", "Response"]))

    @classmethod
    def load_llms_from_file(cls, filename):
        """Load llms from file"""
        with open(filename, "r") as file:
            llms = [line.strip() for line in file.readlines()]
        return cls(llms)

    def get_task_history(self):
        """Get Task history"""
        return self.task_history

    def summary(self):
        """Summary"""
        print("Tasks History:")
        for i, task in enumerate(self.task_history):
            print(f"{i + 1}. {task}")
        print("\nLast Responses:")
        table = [
            [f"LLM {i+1}", response]
            for i, response in enumerate(self.last_responses)
        ]
        print(
            colored(
                tabulate(
                    table,
                    headers=["LLM", "Response"],
                    tablefmt="pretty",
                ),
                "cyan",
            )
        )

    def enable_load_balancing(self):
        """Enable load balancing among LLMs."""
        self.load_balancing = True
        logger.info("Load balancing enabled.")

    def disable_load_balancing(self):
        """Disable load balancing."""
        self.load_balancing = False
        logger.info("Load balancing disabled.")

    async def arun(self, task: str):
        """Asynchronous run the task string"""
        loop = asyncio.get_event_loop()
        futures = [
            loop.run_in_executor(None, lambda llm: llm(task), llm)
            for llm in self.llms
        ]
        for response in await asyncio.gather(*futures):
            print(response)

    def concurrent_run(self, task: str) -> List[str]:
        """Synchronously run the task on all llms and collect responses"""
        try:
            with ThreadPoolExecutor() as executor:
                future_to_llm = {
                    executor.submit(llm, task): llm
                    for llm in self.llms
                }
                responses = []
                for future in as_completed(future_to_llm):
                    try:
                        responses.append(future.result())
                    except Exception as error:
                        print(
                            f"{future_to_llm[future]} generated an"
                            f" exception: {error}"
                        )
            self.last_responses = responses
            self.task_history.append(task)
            return responses
        except Exception as error:
            print(
                f"[ERROR][ModelParallelizer] [ROOT CAUSE] [{error}]"
            )
            raise error

    def add_llm(self, llm: Callable):
        """Add an llm to the god mode"""
        self.llms.append(llm)

    def remove_llm(self, llm: Callable):
        """Remove an llm from the god mode"""
        self.llms.remove(llm)
